/*
 * This file is part of SuperLib, which is an AI Library for OpenTTD
 * Copyright (C) 2010  Leif Linse
 *
 * SuperLib is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * SuperLib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SuperLib; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

 /*
  * Hello AI developers,
  *
  * This main.nut file has been auto-generated. There is no comments about
  * how functions should be used in this file.
  *
  * Please refer to <class_name>.nut files for documentation of the child
  * classes and their functions.
  *
  * Eg for Helper documentation, please refer to helper.nut
  */

class SuperLib
{
	static Helper = class
	{
		function SetSign(tile, message, force_build_sign = false)
		{
			if(!force_build_sign && AIController.GetSetting("debug_signs") != 1)
				return;
		
			local found = false;
			local sign_list = AISignList();
			for(local i = sign_list.Begin(); sign_list.HasNext(); i = sign_list.Next())
			{
				if(AISign.GetLocation(i) == tile)
				{
					if(found)
						AISign.RemoveSign(i);
					else
					{
						if(message == "")
							AISign.RemoveSign(i);
						else
							AISign.SetName(i, message);
						found = true;
					}
				}
			}
		
			if(!found)
				AISign.BuildSign(tile, message);
		}
		function BreakPoint(sign_tile, force_break_point = false)
		{
			if(force_break_point != false)
			{
				if(Helper.HasSign("no_break"))
					return;
		
				if(!Helper.HasSign("break_on"))
				{
					if(AIController.GetSetting("debug_signs") != 1)
						return;
		
				}
			}
		
			AILog.Warning("Break point reached. -> Remove the \"break\" sign to continue.");
			Helper.SetSign(sign_tile, ""); // remove any signs on the tile first
			local sign = AISign.BuildSign(sign_tile, "break");
			while(AISign.IsValidSign(sign)) { AIController.Sleep(1); }
		}
		function HasSign(text)
		{
			local sign_list = AISignList();
			for(local i = sign_list.Begin(); sign_list.HasNext(); i = sign_list.Next())
			{
				if(AISign.GetName(i) == text)
				{
					return true;
				}
			}
			return false;
		}
		function ClearAllSigns()
		{
			local sign_list = AISignList();
			for(local i = sign_list.Begin(); sign_list.HasNext(); i = sign_list.Next())
			{
				AISign.RemoveSign(i);
			}
		}
		function ListValueSum(ai_list)
		{
			local sum = 0;
			local item = ai_list.Begin();
			while(ai_list.HasNext())
			{
				sum += ai_list.GetValue(item);
		
				item = ai_list.Next();
			}
		
			return sum;
		}
		function CopyListSwapValuesAndItems(old_list)
		{
			local new_list = AIList();
			for(local i = old_list.Begin(); old_list.HasNext(); i = old_list.Next())
			{
				local value = old_list.GetValue(i);
				new_list.AddItem(value, i);
			}
		
			return new_list;
		}
		function GetListMinValue(ai_list)
		{
			ai_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // highest last
			return ai_list.GetValue(ai_list.Begin());
		}
		function GetListMaxValue(ai_list)
		{
			ai_list.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest first
			return ai_list.GetValue(ai_list.Begin());
		}
		function SquirrelListToAIList(squirrel_list)
		{
			local ai_list = AIList();
			foreach(item in squirrel_list)
			{
				ai_list.AddItem(item, 0);
			}
		
			return ai_list;
		}
		function GetPAXCargo()
		{
			local cargo_list = AICargoList();
			cargo_list.Valuate(AICargo.HasCargoClass, AICargo.CC_PASSENGERS);
			cargo_list.KeepValue(1);
			cargo_list.Valuate(AICargo.GetTownEffect);
			cargo_list.KeepValue(AICargo.TE_PASSENGERS);
		
			if(!AICargo.IsValidCargo(cargo_list.Begin()))
			{
				AILog.Error("PAX Cargo do not exist");
			}
		
			return cargo_list.Begin();
		}
		function Min(x1, x2)
		{
			return x1 < x2? x1 : x2;
		}
		function Max(x1, x2)
		{
			return x1 > x2? x1 : x2;
		}
		function Clamp(x, min, max)
		{
			x = Helper.Max(x, min);
			x = Helper.Min(x, max);
			return x;
		}
		function Abs(a)
		{
			return a >= 0? a : -a;
		}
	}
	static Tile = class
	{
		function IsStraight(tile1, tile2)
		{
			return AIMap.GetTileX(tile1) == AIMap.GetTileX(tile2) ||
					AIMap.GetTileY(tile1) == AIMap.GetTileY(tile2);
		}
		function GetTileRelative(relative_to_tile, delta_x, delta_y)
		{
			local tile_x = AIMap.GetTileX(relative_to_tile);
			local tile_y = AIMap.GetTileY(relative_to_tile);
		
			local new_x = Helper.Clamp(tile_x + delta_x, 1, AIMap.GetMapSizeX() - 2);
			local new_y = Helper.Clamp(tile_y + delta_y, 1, AIMap.GetMapSizeY() - 2);
		
			return AIMap.GetTileIndex(new_x, new_y);
		}
		function GetNeighbours4MainDir(tile_id)
		{
			local list = AIList();
		
			if(!AIMap.IsValidTile(tile_id))
				return list;
		
			local tile_x = AIMap.GetTileX(tile_id);
			local tile_y = AIMap.GetTileY(tile_id);
		
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_NE), Direction.DIR_NE);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_SE), Direction.DIR_SE);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_SW), Direction.DIR_SW);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_NW), Direction.DIR_NW);
		
			return list;
		}
		function GetNeighbours8(tile_id)
		{
			local list = AIList();
		
			if(!AIMap.IsValidTile(tile_id))
				return list;
		
			local tile_x = AIMap.GetTileX(tile_id);
			local tile_y = AIMap.GetTileY(tile_id);
		
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_N), Direction.DIR_N);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_E), Direction.DIR_E);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_S), Direction.DIR_S);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_W), Direction.DIR_W);
		
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_NE), Direction.DIR_NE);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_SE), Direction.DIR_SE);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_SW), Direction.DIR_SW);
			list.AddItem(Direction.GetAdjacentTileInDirection(tile_id, Direction.DIR_NW), Direction.DIR_NW);
		
			return list;
		}
		function IsAdjacent8ToBuildableTile(tile_id)
		{
			local neighbours = GetNeighbours8(tile_id);
		
			neighbours.Valuate(AITile.IsBuildable);
			neighbours.KeepValue(1);
		
			return !neighbours.IsEmpty();
		}
		function MakeTileRectAroundTile(center_tile, radius)
		{
			local tile_x = AIMap.GetTileX(center_tile);
			local tile_y = AIMap.GetTileY(center_tile);
		
			local x_min = Helper.Clamp(tile_x - radius, 1, AIMap.GetMapSizeX() - 2);
			local x_max = Helper.Clamp(tile_x + radius, 1, AIMap.GetMapSizeX() - 2);
			local y_min = Helper.Clamp(tile_y - radius, 1, AIMap.GetMapSizeY() - 2);
			local y_max = Helper.Clamp(tile_y + radius, 1, AIMap.GetMapSizeY() - 2);
		
			local list = AITileList();
			list.AddRectangle( AIMap.GetTileIndex(x_min, y_min), AIMap.GetTileIndex(x_max, y_max) );
		
			return list;
		}
		function GrowTileRect(tile_list, grow_amount)
		{
			local min_x = AIMap.GetMapSizeX(), min_y = AIMap.GetMapSizeY(), max_x = 0, max_y = 0;
		
			foreach(tile in tile_list)
			{
				local x = AIMap.GetTileX(tile);
				local y = AIMap.GetTileY(tile);
		
				if(x < min_x) min_x = x;
				if(y < min_y) min_y = y;
				if(x > max_x) max_x = x;
				if(y > max_y) max_y = y;
			}
		
			local new_tile_list = AITileList();
		
			// Create the x0,y0 and x1,y1 coordinates for the grown rectangle clamped to the map size (minus a 1 tile border to fully support non-water map borders)
			local x0 = Helper.Max(1, min_x - grow_amount);
			local y0 = Helper.Max(1, min_y - grow_amount);
			local x1 = Helper.Min(AIMap.GetMapSizeX() - 2, min_x + grow_amount);
			local y1 = Helper.Min(AIMap.GetMapSizeY() - 2, min_y + grow_amount);
		
			new_tile_list.AddRectangle(AIMap.GetTileIndex(min_x, min_y), AIMap.GetTileIndex(max_x, max_y));
			return new_tile_list;
		}
		function IsBuildOnSlopeEnabled()
		{
			if (!AIGameSettings.IsValid("build_on_slopes"))
			{
				AILog.Error("Game setting \"build_on_slopes\" is not valid anymore!");
				KABOOOOOOOM_game_setting_is_not_valid_anymore // Make sure this error is found!
			}
		
			return AIGameSettings.GetValue("build_on_slopes") != false;
		}
		function IsDownSlope(tile_id, direction)
		{
			local opposite_dir =  Direction.TurnDirClockwise45Deg(direction, 4);
			return Tile.IsUpSlope(tile_id, opposite_dir);
		}
		function IsUpSlope(tile_id, direction)
		{
			local slope = AITile.GetSlope(tile_id);
		
			switch(direction)
			{
				case Direction.DIR_NE:
					return slope == AITile.SLOPE_NE; // Has N & E corners raised
		
				case Direction.DIR_SE:
					return slope == AITile.SLOPE_SE;
		
				case Direction.DIR_SW:
					return slope == AITile.SLOPE_SW;
		
				case Direction.DIR_NW:
					return slope == AITile.SLOPE_NW;
		
				default:
					return false;
			}
		
			return false;
		}
		function IsBuildOnSlope_DownSlope(tile_id, direction)
		{
			local opposite_dir = Direction.TurnDirClockwise45Deg(direction, 4);
			return Tile.IsBuildOnSlope_UpSlope(tile_id, opposite_dir);
		}
		function IsBuildOnSlope_UpSlope(tile_id, direction)
		{
			// If build on slopes is disabled, then call IsUpSlope instead
			if (!Tile.IsBuildOnSlopeEnabled())
			{
				return Tile.IsUpSlope(tile_id, direction);
			}
		
			local slope = AITile.GetSlope(tile_id);
		
			switch(direction)
			{
				case Direction.DIR_NE:
					return ((slope & AITile.SLOPE_N) != 0x00 || (slope & AITile.SLOPE_E) != 0x00) && // must have either N or E tile raised
							((slope & AITile.SLOPE_S) == 0x00 && (slope & AITile.SLOPE_W) == 0x00); // and neither of S or W
		
				case Direction.DIR_SE:
					return ((slope & AITile.SLOPE_S) != 0x00 || (slope & AITile.SLOPE_E) != 0x00) && 
							((slope & AITile.SLOPE_N) == 0x00 && (slope & AITile.SLOPE_W) == 0x00);
		
				case Direction.DIR_SW:
					return ((slope & AITile.SLOPE_S) != 0x00 || (slope & AITile.SLOPE_W) != 0x00) && 
							((slope & AITile.SLOPE_N) == 0x00 && (slope & AITile.SLOPE_E) == 0x00);
		
				case Direction.DIR_NW:
					return ((slope & AITile.SLOPE_N) != 0x00 || (slope & AITile.SLOPE_W) != 0x00) && 
							((slope & AITile.SLOPE_S) == 0x00 && (slope & AITile.SLOPE_E) == 0x00);
		
				default:
					return false;
			}
		
			return false;
		}
		function IsBuildOnSlope_Flat(tile_id)
		{
			local slope = AITile.GetSlope(tile_id);
		
			if (slope == AITile.SLOPE_FLAT) return true;
		
			// Only accept three raised corner tiles if build on slope is enabled
			if (!Tile.IsBuildOnSlopeEnabled())
			{
				return false;
			}
		
			// If two opposite corners are raised -> return true, else false
			return ((slope & AITile.SLOPE_N) != 0x00 && (slope & AITile.SLOPE_S) != 0x00) ||
					((slope & AITile.SLOPE_E) != 0x00 && (slope & AITile.SLOPE_W) != 0x00);
		
		}
	}
	static Direction = class
	{
		static DIR_N = 0;
		static DIR_NE = 1;
		static DIR_SE = 3;
		static DIR_SW = 5;
		static DIR_NW = 7;
		static DIR_FIRST = 0;
		static DIR_LAST = 7;
		static DIR_INVALID = 8;
		function GetDirString(dir)
		{
			switch(dir)
			{
				case Direction.DIR_N:
					return "N";
				case Direction.DIR_NE:
					return "NE";
				case Direction.DIR_E:
					return "E";
				case Direction.DIR_SE:
					return "SE";
				case Direction.DIR_S:
					return "S";
				case Direction.DIR_SW:
					return "SW";
				case Direction.DIR_W:
					return "W";
				case Direction.DIR_NW:
					return "NW";
				default:
					AILog.Error("Direction::GetDirString: invalid direction: " + in_direction);
					return -1;
			}
			
		}
		function IsMainDir(dir)
		{
			return dir == Direction.DIR_NE ||
					dir == Direction.DIR_SE ||
					dir == Direction.DIR_SW ||
					dir == Direction.DIR_NW;
		}
		function IsDiagonalDir(dir)
		{
			return dir == Direction.DIR_N ||
					dir == Direction.DIR_E ||
					dir == Direction.DIR_S ||
					dir == Direction.DIR_W;
		}
		function GetAdjacentTileInDirection(from_tile, in_direction)
		{
			switch(in_direction)
			{
				case Direction.DIR_N:
					return Tile.GetTileRelative(from_tile, -1, -1);
				case Direction.DIR_NE:
					return Tile.GetTileRelative(from_tile, -1,  0);
				case Direction.DIR_E:
					return Tile.GetTileRelative(from_tile, -1,  1);
				case Direction.DIR_SE:
					return Tile.GetTileRelative(from_tile,  0,  1);
				case Direction.DIR_S:
					return Tile.GetTileRelative(from_tile,  1,  1);
				case Direction.DIR_SW:
					return Tile.GetTileRelative(from_tile,  1,  0);
				case Direction.DIR_W:
					return Tile.GetTileRelative(from_tile,  1, -1);
				case Direction.DIR_NW:
					return Tile.GetTileRelative(from_tile,  0, -1);
				default:
					AILog.Error("Direction::GetAdjacentTileInDirection: invalid direction: " + in_direction);
					return -1;
			}
		}
		function GetDirectionToAdjacentTile(tile, adjacent_tile)
		{
			local rel_x = AIMap.GetTileX(adjacent_tile) - AIMap.GetTileX(tile);
			local rel_y = AIMap.GetTileY(adjacent_tile) - AIMap.GetTileY(tile);
		
			local rel_dir = [ [Direction.DIR_N, [-1, -1]],
					[Direction.DIR_NE, [-1,  0]],
					[Direction.DIR_E,  [-1,  1]],
					[Direction.DIR_SE, [ 0,  1]],
					[Direction.DIR_S,  [ 1,  1]],
					[Direction.DIR_SW, [ 1,  0]],
					[Direction.DIR_W,  [ 1, -1]],
					[Direction.DIR_NW, [ 0, -1]] ];
		
			foreach(dir in rel_dir)
			{
				local rel = dir[1];
				local X = 0;
				local Y = 1;
		
				if(rel[X] == rel_x && rel[Y] == rel_y)
				{
					return dir[0];
				}
			}
		
			return Direction.DIR_INVALID;
		}
		function GetDirectionToTile(tile1, tile2)
		{
			local rel_x = AIMap.GetTileX(tile1) - AIMap.GetTileX(tile2);
			local rel_y = AIMap.GetTileY(tile1) - AIMap.GetTileY(tile2);
		
			if(rel_x == 0 && rel_y == 0)
			{
				return Direction.DIR_INVALID;
			}
			else if (Helper.Abs(rel_x) >= 1 && Helper.Abs(rel_y) >= 1)
			{
				// Neither of NE, NW, SW, SE
		
				if (Helper.Abs(rel_x) == Helper.Abs(rel_y))
				{
					// Same amplitude of rel_x and rel_y => N, W, S or E.
					if(rel_x <= -1 && rel_y <= -1) return Direction.DIR_N;
					if(rel_x <= -1 && rel_y >= 1) return Direction.DIR_E;
					if(rel_x >= 1 && rel_y >= 1) return Direction.DIR_S;
					if(rel_x >= 1 && rel_y <= -1) return Direction.DIR_S;
		
					// Error
					return Direction.DIR_INVALID;
				}
				else
				{
					// Not a valid direction
					return Direction.DIR_INVALID;
				}
			}
			else
			{
				// either rel_x or rel_y is 0 and the other has a value > 0.
				
				if (rel_x < 0) return Direction.DIR_NE;
				if (rel_y > 0) return Direction.DIR_SE;
				if (rel_x > 0) return Direction.DIR_SW;
				if (rel_y < 0) return Direction.DIR_NW;
		
				// Error
				return Direction.DIR_INVALID
			}
				
			// Error
			return Direction.DIR_INVALID
		}
		function TurnDirClockwise45Deg(dir, num_45_deg)
		{
			/* Make sure to only handle positive turns */
			if(num_45_deg < 0) return TurnDirAntiClockwise45Deg(dir, -num_45_deg);
		
			dir += num_45_deg
			
			while(dir > Direction.DIR_LAST)
			{
				dir -= (Direction.DIR_LAST - Direction.DIR_FIRST) + 1;
			} 
		
			return dir;
		}
		function TurnDirAntiClockwise45Deg(dir, num_45_deg)
		{
			/* Make sure to only handle positive turns */
			if(num_45_deg < 0) return TurnDirClockwise45Deg(dir, -num_45_deg);
		
			dir -= num_45_deg
			
			while(dir < Direction.DIR_FIRST)
			{
				dir += (Direction.DIR_LAST - Direction.DIR_FIRST) + 1;
			}
		
			return dir;
		}
	}
	}
