#!ruby
# 
# This file is part of SuperLib, which is an AI Library for OpenTTD
# Copyright (C) 2010  Leif Linse
# 
# SuperLib is free software; you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License
# 
# SuperLib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SuperLib; If not, see <http://www.gnu.org/licenses/> or
# write to the Free Software Foundation, Inc., 51 Franklin Street, 
# Fifth Floor, Boston, MA 02110-1301 USA.
# 


# --- functions ----

def SkipUntil(file, line_contains)
	
	while(line = file.gets)
		if line.match(line_contains)
			return
		end
	end
end

def ReadClassDeclaration(nut_file, main_file, sub_lib)

	# Skip pass all comments etc. until the class declaration is found
	SkipUntil(nut_file, "class " + sub_lib)

	# Loop through all lines inside the class and look for static "constants"
	while(line = nut_file.gets)

		# Add lines to main file if they contain a static constant
		# eg
		# static SomeName_98 = ...
		if line.match(/static [a-zA-Z0-9_]*\s=/) != nil
			main_file.puts("\t\t" + line.gsub(/^\s/, ""))
		end

		# Stop if end of class declaration is reached
		if line.match(/^\}/) != nil
			return
		end
	end
end

def ReadFunctionDefinitions(nut_file, main_file, sub_lib)
	while (true)
		# find next function
		line = ""
		while (line.match(/^function/) == nil) 
			line = nut_file.gets; 
			if line == nil
				return
			end
		end

		# line is now a function definition line
		func_decl = line.gsub(/^function [a-zA-Z]*::/, "function "); # remove the class name
		main_file.puts("\t\t" + func_decl)

		while (line.match(/^\}/) == nil) # stops after the ending } has been put to the main file
			line = nut_file.gets 
			main_file.puts("\t\t" + line)
		end
	end
	
end

# --- sub libs ----

# Add the names of the sub libraries here. 
# Helper is expected to be found in helper.nut etc.
sub_libs = ["Helper", "Tile", "Direction"]


# --- main ----

main_nut_file = File.new("main.nut", "w")

main_nut_file.puts("/*
 * This file is part of SuperLib, which is an AI Library for OpenTTD
 * Copyright (C) 2010  Leif Linse
 *
 * SuperLib is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * SuperLib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SuperLib; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

 /*
  * Hello AI developers,
  *
  * This main.nut file has been auto-generated. There is no comments about
  * how functions should be used in this file.
  *
  * Please refer to <class_name>.nut files for documentation of the child
  * classes and their functions.
  *
  * Eg for Helper documentation, please refer to helper.nut
  */

class SuperLib
{
")

sub_libs.each { |sub_lib|

	nut_file = File.new(sub_lib.downcase + ".nut") 

	main_nut_file.puts("\tstatic " + sub_lib + " = class")
	main_nut_file.puts("\t{")

	# Read the class declaration
	ReadClassDeclaration(nut_file, main_nut_file, sub_lib)

	# Read the class function definitions
	ReadFunctionDefinitions(nut_file, main_nut_file, sub_lib)

	main_nut_file.puts("\t}") # close child class
}

main_nut_file.puts("\t}") # close main class

main_nut_file.close

# --- Create bundle ---

if not ARGV.empty? and ARGV[0] == "bundle"

	system("mkdir SuperLib")
	system("cp *.nut SuperLib")
	system("tar -cf SuperLib.tar SuperLib")

	system("rm -rf SuperLib")
end
